#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <array>
#include <vector>
#ifdef __APPLE__
#include <GL/glew.h>
#else
#include <OpenGL/glew.h>
#endif
#include <GLUT/glut.h>
#include <glm/glm.hpp>
#include <glm/ext.hpp>
#include <glm/gtc/type_ptr.hpp>

#pragma clang diagnostic ignored "-Wdeprecated-declarations"

using namespace std;

GLuint pyramidVao, lightVao;
GLuint gradientProgId, solidProgId;
glm::vec3 camPos   = glm::vec3(0.0f, 0.0f,  -5.0f);
glm::vec3 camFront = glm::vec3(0.0f, 0.0f, 1.0f);
glm::vec3 camUp    = glm::vec3(0.0f, 1.0f,  0.0f);
float yaw = 1.57, pitch = 0;
bool doScale, doRotate, doTranslate;

void setProjectionAndViewUniforms(GLuint progId) {
	GLuint projId = glGetUniformLocation(progId, "projection");
	glm::mat4 proj = glm::perspective(glm::radians(45.f), 1.33f, 0.01f, 10000.f);
	glUniformMatrix4fv(projId, 1, GL_FALSE, glm::value_ptr(proj));

	GLuint viewId = glGetUniformLocation(progId, "view");
	glm::mat4 view = glm::lookAt(camPos, camPos + camFront, camUp);
	glUniformMatrix4fv(viewId, 1, GL_FALSE, glm::value_ptr(view));
}

void drawLight(float d, glm::vec3 lightPos) {
  	glUseProgram(solidProgId);
	glBindVertexArray(lightVao);
	setProjectionAndViewUniforms(solidProgId);
	glm::mat4 model = glm::translate(glm::mat4(1.f), lightPos);
	model = glm::scale(model, glm::vec3(0.2));
	GLuint modelId = glGetUniformLocation(solidProgId, "model");
	glUniformMatrix4fv(modelId, 1, GL_FALSE, glm::value_ptr(model));

	GLuint colorLoc = glGetUniformLocation(solidProgId, "color");
	glm::vec3 color = glm::vec3(1, 1, 1);
	glUniform3fv(colorLoc, 1, glm::value_ptr(color));
		
	glDrawArrays(GL_TRIANGLES, 0, 36);
};

void drawPyramids(float d, glm::vec3 lightPos) {
	glUseProgram(gradientProgId);
	glBindVertexArray(pyramidVao);
	setProjectionAndViewUniforms(gradientProgId);

	GLuint lightPosLoc = glGetUniformLocation(gradientProgId, "lightPos");
	glUniform3fv(lightPosLoc, 1, glm::value_ptr(lightPos));
	
	GLuint viewPosLoc = glGetUniformLocation(gradientProgId, "viewPos");
	glUniform3fv(viewPosLoc, 1, glm::value_ptr(camPos));

	GLuint modelId = glGetUniformLocation(gradientProgId, "model");

	for (int i = 0; i < 10; i++) {

		glm::mat4 model = glm::mat4(1.f);

		model = glm::translate(model, glm::vec3(sin(i * 30) * 10, 0, i * 2 - 10));
		
		if (doRotate) {
			model = glm::rotate(model, d * glm::radians(30.f), glm::vec3(0.f, 1.f, 0.f));
			model = glm::rotate(model, d * glm::radians(20.f), glm::vec3(1.f, 0.f, 0.f));
		}
		
		if (doScale)
			model = glm::scale(model, glm::vec3(1.f, 0.7f + 0.7f * (1 + sin(d + (i + 3))), 1.f));

		if (doTranslate)
			model = glm::translate(model, glm::vec3(sin(d + (i + 1)), cos(d + (i + -3)), sin(d + (i + 4))));
	

		glUniformMatrix4fv(modelId, 1, GL_FALSE, glm::value_ptr(model));
		
		glDrawArrays(GL_TRIANGLES, 0, 12);
	}

};

void display() {
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	float d = (float)glutGet(GLUT_ELAPSED_TIME) * 0.001f;

	glm::vec3 lightPos = glm::vec3(sin(d) * 10, 0, cos(d) * 10);

	drawPyramids(d, lightPos);
	drawLight(d, lightPos);

	glutSwapBuffers();
}

void attachShader(GLuint progId, const char* filePath, GLenum type) {
	GLuint shader = glCreateShader(type);

	if (!shader) {
		fprintf(stderr, "error creating shader\n");
		exit(1);
	}

	ifstream file(filePath);
	stringstream buffer;
	buffer << file.rdbuf();
	string str = buffer.str();
	const char* contents = str.c_str();

	glShaderSource(shader, 1, (const GLchar**)&contents, NULL);
	glCompileShader(shader);
	GLint success;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &success);
	if (!success) {
		GLchar log[1024];
		glGetShaderInfoLog(shader, 1024, NULL, log);
		fprintf(stderr, "error: %s\n", log);
		exit(1);
	}
	glAttachShader(progId, shader);
}

GLuint compileShaders(char* vertexShader, char* fragmentShader) {
	GLuint progId = glCreateProgram();

	attachShader(progId, vertexShader, GL_VERTEX_SHADER);
	attachShader(progId, fragmentShader, GL_FRAGMENT_SHADER);

	glLinkProgram(progId);
	GLint success = 0;
	glGetProgramiv(progId, GL_LINK_STATUS, &success);
	if (!success) {
		GLchar log[1024];
		glGetProgramInfoLog(progId, sizeof(log), NULL, log);
		fprintf(stderr, "error linking: %s\n", log);
		exit(1);
	}

	return progId;
}

#define BUFFER_OFFSET(i) ((char *)NULL + (i))

GLuint setupBuffers(glm::vec3* vertices, glm::vec3* normals, GLuint progId) {

	GLfloat colors[] = {
		0, 1, 0, 1,
		1, 0, 0, 1,
		0, 0, 1, 1,

		0, 1, 0, 1,
		1, 0, 0, 1,
		0, 0, 1, 1,

		0, 1, 0, 1,
		1, 0, 0, 1,
		0, 0, 1, 1,

		0, 1, 0, 1,
		1, 0, 0, 1,
		0, 0, 1, 1
	};

	GLuint numVerts = 12;

	GLuint vbo;
	glGenBuffers(1, &vbo);

	GLuint vao;
	glGenVertexArrays(1, &vao);

	GLuint posId = glGetAttribLocation(progId, "vPosition");
	GLuint colorId = glGetAttribLocation(progId, "vColor");
	GLuint normalLoc = glGetAttribLocation(progId, "vNormal");

	GLuint vertsLen = numVerts * 3 * sizeof(GLfloat);
	GLuint colorsLen = numVerts * 4 * sizeof(GLfloat);
	GLuint normalLen = numVerts * 3 * sizeof(GLfloat);

	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, vertsLen + colorsLen + normalLen, NULL, GL_STATIC_DRAW);

	glBufferSubData(GL_ARRAY_BUFFER, 0, vertsLen, glm::value_ptr(vertices[0]));
	glBufferSubData(GL_ARRAY_BUFFER, vertsLen, colorsLen, colors);
	glBufferSubData(GL_ARRAY_BUFFER, vertsLen + colorsLen, normalLen, glm::value_ptr(normals[0]));
	
	glBindVertexArray(vao);

	glEnableVertexAttribArray(posId);
	glVertexAttribPointer(posId, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glEnableVertexAttribArray(colorId);
	glVertexAttribPointer(colorId, 4, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(vertsLen));

	glEnableVertexAttribArray(normalLoc);
	glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(vertsLen + colorsLen));

	return vao;
}

vector<glm::vec3> quadToTriangles(glm::vec3 *quads) {
	vector<glm::vec3> triangles(6);
	triangles[0] = quads[0];
	triangles[1] = quads[1];
	triangles[2] = quads[2];
	triangles[3] = quads[2];
	triangles[4] = quads[3];
	triangles[5] = quads[0];
	return triangles;
}

template <typename T>
void append(vector<T> &a, vector<T> &b) {
	a.insert(a.end(), b.begin(), b.end());
};

void setupLightBuffers(GLuint progId) {
  	vector<glm::vec3> vertices;
	glm::vec3 front[] = {
		glm::vec3(1, -1, -1),
		glm::vec3(-1, -1, -1),
		glm::vec3(-1, 1, -1),
		glm::vec3(1, 1, -1)
	};
	vector<glm::vec3> frontTriangles = quadToTriangles(front);
	append(vertices, frontTriangles);

	glm::vec3 back[] = {
		glm::vec3(1, 1, 1),
		glm::vec3(-1, 1, 1),
		glm::vec3(-1, -1, 1),
		glm::vec3(1, -1, 1)
	};
	vector<glm::vec3> backQuads = quadToTriangles(back);
	append(vertices, backQuads);

	glm::vec3 top[] = {
		glm::vec3(1, 1, -1),
		glm::vec3(-1, 1, -1),
		glm::vec3(-1, 1, 1),
		glm::vec3(1, 1, 1)
	};
	vector<glm::vec3> topTriangles = quadToTriangles(top);
	append(vertices, topTriangles);

	glm::vec3 bottom[] = {
		glm::vec3(1, -1, 1),
		glm::vec3(-1, -1, 1),
		glm::vec3(-1, -1, -1),
		glm::vec3(1, -1, -1)
	};
	vector<glm::vec3> bottomTriangles = quadToTriangles(bottom);
	append(vertices, bottomTriangles);

	glm::vec3 left[] = {
		glm::vec3(-1, 1, 1),
		glm::vec3(-1, 1, -1),
		glm::vec3(-1, -1, -1),
		glm::vec3(-1, -1, 1)
	};
	vector<glm::vec3> leftTriangles = quadToTriangles(left);
	append(vertices, leftTriangles);

	glm::vec3 right[] = {
		glm::vec3(1, 1, -1),
		glm::vec3(1, 1, 1),
		glm::vec3(1, -1, 1),
		glm::vec3(1, -1, -1)
	};
	vector<glm::vec3> rightTriangles = quadToTriangles(right);
	append(vertices, rightTriangles);
	GLuint verticesSize = 36 * 3 * sizeof(GLfloat);

	glGenVertexArrays(1, &lightVao);
	GLuint vbo;
	glBindVertexArray(lightVao);
	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, verticesSize, NULL, GL_STATIC_DRAW);
	glBufferSubData(GL_ARRAY_BUFFER, 0, verticesSize, glm::value_ptr(vertices[0]));
	GLuint posLoc = glGetAttribLocation(progId, "vPosition");
	glEnableVertexAttribArray(posLoc);
	glVertexAttribPointer(posLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);
}

void validateProgram(GLuint progId) {
	glValidateProgram(progId);
	
	GLint success;
	glGetProgramiv(progId, GL_VALIDATE_STATUS, &success);
	if (!success) {
		GLchar log[1024];
		glGetProgramInfoLog(progId, sizeof(log), NULL, log);
		fprintf(stderr, "error: %s\n", log);
		exit(1);
	}
}

void init() {
	glm::vec3 vertices[12] = {
		glm::vec3(0.0f, 1.0f, 0.0f),
		glm::vec3(1.0f, -1.0f, -1.0f),
		glm::vec3(-1.0f, -1.0f, -1.0f),

		glm::vec3(0.0f, 1.0f, 0.0f),
		glm::vec3(-1.0f, -1.0f, 1.0f),
		glm::vec3(1.0f, -1.0f, 1.0f),

		glm::vec3(0.0f, 1.0f, 0.0f),
		glm::vec3(-1.0f, -1.0f, -1.0f),
		glm::vec3(-1.0f, -1.0f, 1.0f),

		glm::vec3(0.0f, 1.0f, 0.0f),
		glm::vec3(1.0f, -1.0f, 1.0f),
		glm::vec3(1.0f, -1.0f, -1.0f)
	};

	// work out the normals
	glm::vec3 normals[12];
	for (int i = 0; i < 4; i++) {
		glm::vec3 a = vertices[i * 3];
		glm::vec3 b = vertices[i * 3 + 1];
		glm::vec3 c = vertices[i * 3 + 2];
		glm::vec3 u = glm::normalize(a - c);
	  	glm::vec3 v = glm::normalize(b - c);
		glm::vec3 norm = glm::normalize(glm::cross(v, u));
		for(int j = 0; j < 3; j++) {
			normals[i * 3 + j] = glm::vec3(norm);
		}
	}

	gradientProgId = compileShaders((char*)"vertex.glsl", (char*)"fragment.glsl");
	glUseProgram(gradientProgId);
	pyramidVao = setupBuffers(vertices, normals, gradientProgId);
	validateProgram(gradientProgId);

	solidProgId = compileShaders((char*)"solidvertex.glsl", (char*)"solidfrag.glsl");
	glUseProgram(solidProgId);
	setupLightBuffers(solidProgId);
	validateProgram(solidProgId);

	glEnable(GL_DEPTH_TEST); 
	glEnable(GL_CULL_FACE); 
}

bool* keyStates = new bool[256];

void keyboard(unsigned char key, int x, int y) {
	keyStates[key] = true;
	if (key == 'z')
		doScale = !doScale;
	if (key == 'x')
		doRotate = !doRotate;
	if (key == 'c')
		doTranslate = !doTranslate;
}

void keyboardUp(unsigned char key, int x, int y) {
	keyStates[key] = false;
}

void timer(int _) {
	float xSpeed = 0.f, ySpeed = 0.f, zSpeed = 0.f;
	if (keyStates['w'])
		zSpeed = 0.1f;
	if (keyStates['s'])
		zSpeed = -0.1f;
	if (keyStates['a'])
		xSpeed = 0.1f;
	if (keyStates['d'])
		xSpeed = -0.1f;
	if (keyStates['q'])
		ySpeed = 0.1f;
	if (keyStates['e'])
		ySpeed = -0.1f;

	camPos.x += xSpeed * sin(yaw) + zSpeed * cos(yaw);
	camPos.y += ySpeed;
	camPos.z += zSpeed * sin(yaw) - xSpeed * cos(yaw);
	glutPostRedisplay();
	glutTimerFunc(16, timer, 0);
}

int prevMouseX, prevMouseY;
bool firstMouse = true;

void motion(int x, int y) {
	if (firstMouse) {
		prevMouseX = x;
		prevMouseY = y;
		firstMouse = false;
	}
	int dx = x - prevMouseX, dy = y - prevMouseY;

	prevMouseX = x;
	prevMouseY = y;

	const float sensitivity = 0.005f;
	yaw += dx * sensitivity;
	pitch -= dy * sensitivity;

	glm::vec3 front;
	front.x = cos(pitch) * cos(yaw);
	front.y = sin(pitch);
	front.z = cos(pitch) * sin(yaw);
	camFront = glm::normalize(front);
}

void mouse(int button, int state, int x, int y) {
	if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
		firstMouse = true;
}

int main(int argc, char** argv) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGB|GLUT_3_2_CORE_PROFILE);
	glutInitWindowSize(800, 600);
	int win = glutCreateWindow("Hello Triangle");
	glutDisplayFunc(display);

	glewInit();
	
	init();

	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutTimerFunc(16, timer, 0);
	glutMotionFunc(motion);
	glutMouseFunc(mouse);

	glutMainLoop();

	return 0;
}

